#include "OpenGL_win32.h"
/*
#define VC_EXTRALEAN
#define WIN32_LEAN_AND_MEAN
#undef HDC
#undef HGLRC
#undef APIENTRY
#undef WINGDIAPI
#include <windows.h>
*/
#include <stdio.h>
#include "wglext.h"

static HDC currentDC;
static HGLRC currentHGLRC;

HMODULE hCg;

//--------------------------------------------
// Cg bindings
//--------------------------------------------

typedef void* CGcontext;
typedef void* CGerror;
typedef struct cg_profile* CGprofile;
typedef struct cg_program* CGprogram;

// 0x1010, 0x1011
#define CG_SOURCE 0x1010
#define CG_COMPILED_PROGRAM 0x100a

typedef CGcontext (*PFNCGCREATECONTEXTPROC)();
typedef void (*PFNCGDESTROYCONTEXTPROC)(CGcontext);
typedef CGprofile (*PFNCGGETPROFILEPROC)(const char*);

// flag is 0x1010 or 0x1011
typedef CGprogram (*PFNCGCREATEPROGRAMPROC)(CGcontext, int flag, const char*, CGprofile, int, int);
typedef void (*PFNCGCOMPILEPROGRAMPROC)(CGprogram);

// flag is 0x1008, 0x1009, 0x100a, 0x100b
typedef char* (*PFNCGGETPROGRAMSTRINGPROC)(CGprogram, int flag);
typedef char* (*PFNCGGETLASTERRORSTRINGPROC)(CGerror*);

PFNCGCREATECONTEXTPROC cgCreateContext = NULL;
PFNCGDESTROYCONTEXTPROC cgDestroyContext = NULL;
PFNCGGETPROFILEPROC cgGetProfile = NULL;
PFNCGCREATEPROGRAMPROC cgCreateProgram = NULL;
PFNCGCOMPILEPROGRAMPROC cgCompileProgram = NULL;
PFNCGGETPROGRAMSTRINGPROC cgGetProgramString = NULL;
PFNCGGETLASTERRORSTRINGPROC cgGetLastErrorString = NULL;

void* getCgProcAddress(char* name)
{
   void *ptr= NULL;

   ptr= (void*)GetProcAddress(hCg, name);

   if (!ptr) {
	   char err[256];
	   sprintf(err, "Cg function not available: %s\n", name);
	   MessageBox(NULL, err, "Error", MB_OK);
	   osDebugBreak();
   }

   return ptr;
}

static CGcontext cgCtx;

void gfxCgInit() {
	hCg = LoadLibrary("cg.dll");

	cgCreateContext = (PFNCGCREATECONTEXTPROC) getCgProcAddress("cgCreateContext");
	cgDestroyContext = (PFNCGDESTROYCONTEXTPROC) getCgProcAddress("cgDestroyContext");
	cgGetProfile = (PFNCGGETPROFILEPROC) getCgProcAddress("cgGetProfile");
	cgCreateProgram = (PFNCGCREATEPROGRAMPROC) getCgProcAddress("cgCreateProgram");
	cgCompileProgram = (PFNCGCOMPILEPROGRAMPROC) getCgProcAddress("cgCompileProgram");
	cgGetProgramString = (PFNCGGETPROGRAMSTRINGPROC) getCgProcAddress("cgGetProgramString");
	cgGetLastErrorString = (PFNCGGETLASTERRORSTRINGPROC) getCgProcAddress("cgGetLastErrorString");

	cgCtx = cgCreateContext();
}

void gfxCgShutdown() {
	cgDestroyContext(cgCtx);
	FreeLibrary(hCg);
}

const char* gfxCgCompileShader(const char* source, const char* profile) {
	const char* errstr = NULL;
	CGerror error;

	CGprofile prof = cgGetProfile(profile);
	CGprogram program = cgCreateProgram(cgCtx, CG_SOURCE, source, prof, 0, 0);
	errstr = cgGetLastErrorString(&error);
	if (!program || errstr) {
		char buf[1024];
		sprintf(buf, "cgCreateProgram failed:\n'%s'\n\n%s\n", errstr, source);
		MessageBox(NULL, buf, "Error", MB_OK);
		return NULL;
	}
	cgCompileProgram(program);
	errstr = cgGetLastErrorString(&error);
	if (errstr) {
		char buf[1024];
		sprintf(buf, "cgCompileProgram failed:\n'%s'\n\n%s\n", errstr, source);
		MessageBox(NULL, buf, "Error", MB_OK);
		return NULL;
	}

	const char* out = cgGetProgramString(program, CG_COMPILED_PROGRAM);
	if (!out) {
		char buf[1024];
		sprintf(buf, "cgGetProgramString failed.\n\n%s\n", source);
		MessageBox(NULL, buf, "Error", MB_OK);
		return NULL;
	}
	return out;
}

//--------------------------------------------
// GL extensions
//--------------------------------------------

PFNGLDRAWRANGEELEMENTSPROC	 glDrawRangeElementsARB = NULL;;
PFNGLCREATESHADERPROC glCreateShader = NULL;;
PFNGLVERTEXATTRIBPOINTERPROC glVertexAttribPointer = NULL;;
PFNGLVERTEXATTRIB4FPROC glVertexAttrib4f = NULL;
PFNGLENABLEVERTEXATTRIBARRAYPROC glEnableVertexAttribArray = NULL;;
PFNGLDISABLEVERTEXATTRIBARRAYPROC glDisableVertexAttribArray = NULL;;

// multi-texture
PFNGLACTIVETEXTUREARBPROC    glActiveTextureARB= NULL;
PFNGLCLIENTACTIVETEXTUREPROC glClientActiveTextureARB= NULL;
PFNGLMULTITEXCOORD2FPROC     glMultiTexCoord2fARB= NULL;
PFNGLMULTITEXCOORD3FPROC     glMultiTexCoord3fARB= NULL;

// vertex buffer objects
PFNGLGENBUFFERSARBPROC    glGenBuffersARB = NULL;
PFNGLBINDBUFFERARBPROC    glBindBufferARB = NULL;
PFNGLBUFFERDATAARBPROC    glBufferDataARB = NULL;
PFNGLDELETEBUFFERSARBPROC glDeleteBuffersARB = NULL;
PFNGLMAPBUFFERPROC        glMapBufferARB = NULL;
PFNGLUNMAPBUFFERPROC      glUnmapBufferARB= NULL;

// ARB_vertex_program + ARB_fragment_program
PFNGLGENPROGRAMSARBPROC					glGenProgramsARB;
PFNGLPROGRAMSTRINGARBPROC				glProgramStringARB;
PFNGLBINDPROGRAMARBPROC					glBindProgramARB;
PFNGLDELETEPROGRAMSARBPROC				glDeleteProgramsARB;
PFNGLPROGRAMENVPARAMETER4FARBPROC		glProgramEnvParameter4fARB;
PFNGLPROGRAMENVPARAMETER4FVARBPROC		glProgramEnvParameter4fvARB;
PFNGLPROGRAMLOCALPARAMETER4FARBPROC		glProgramLocalParameter4fARB;
PFNGLPROGRAMLOCALPARAMETER4FVARBPROC	glProgramLocalParameter4fvARB;

// GLSL shader stuff
PFNGLCREATESHADEROBJECTARBPROC   glCreateShaderObjectARB = NULL;
PFNGLSHADERSOURCEARBPROC         glShaderSourceARB = NULL;
PFNGLCOMPILESHADERARBPROC        glCompileShaderARB = NULL;
PFNGLCREATEPROGRAMOBJECTARBPROC  glCreateProgramObjectARB = NULL;
PFNGLATTACHOBJECTARBPROC         glAttachObjectARB = NULL;
PFNGLLINKPROGRAMARBPROC          glLinkProgramARB = NULL;
PFNGLUSEPROGRAMOBJECTARBPROC     glUseProgramObjectARB = NULL;
PFNGLGETOBJECTPARAMETERFVARBPROC glGetObjectParameterfvARB = NULL;
PFNGLGETOBJECTPARAMETERIVARBPROC glGetObjectParameterivARB = NULL;
PFNGLGETINFOLOGARBPROC           glGetInfoLogARB= NULL;
PFNGLGETUNIFORMLOCATIONARBPROC   glGetUniformLocationARB = NULL;
PFNGLUNIFORM1IARBPROC            glUniform1iARB = NULL;
PFNGLUNIFORM1FARBPROC            glUniform1fARB = NULL;
PFNGLUNIFORM2FARBPROC            glUniform2fARB = NULL;
PFNGLUNIFORM3FARBPROC            glUniform3fARB = NULL;
PFNGLUNIFORM4FARBPROC            glUniform4fARB = NULL;
PFNGLUNIFORM1FVARBPROC           glUniform1fvARB = NULL;
PFNGLUNIFORM2FVARBPROC           glUniform2fvARB = NULL;
PFNGLUNIFORM3FVARBPROC           glUniform3fvARB = NULL;
PFNGLUNIFORM4FVARBPROC           glUniform4fvARB = NULL;

PFNGLACTIVESTENCILFACEEXTPROC    glActiveStencilFaceEXT = NULL;
PFNGLSTENCILOPSEPARATEATIPROC    glStencilOpSeparateATI= NULL;
PFNGLSTENCILFUNCSEPARATEATIPROC  glStencilFuncSeparateATI= NULL;

// fbo
PFNGLGENFRAMEBUFFERSEXTPROC      glGenFramebuffersEXT = NULL;
PFNGLBINDFRAMEBUFFEREXTPROC      glBindFramebufferEXT = NULL;
PFNGLGENRENDERBUFFERSEXTPROC     glGenRenderbuffersEXT = NULL;
PFNGLRENDERBUFFERSTORAGEEXTPROC  glRenderbufferStorageEXT= NULL;
PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbufferEXT= NULL;
PFNGLFRAMEBUFFERTEXTURE2DEXTPROC glFramebufferTexture2DEXT= NULL;
PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC glCheckFramebufferStatusEXT= NULL;
PFNGLBINDRENDERBUFFEREXTPROC     glBindRenderbufferEXT= NULL;

// pbuffer
static PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = NULL;
static PFNWGLCREATEPBUFFERARBPROC wglCreatePbufferARB = NULL;
static PFNWGLGETPBUFFERDCARBPROC wglGetPbufferDCARB = NULL;
static PFNWGLRELEASEPBUFFERDCARBPROC wglReleasePbufferDCARB = NULL;
static PFNWGLDESTROYPBUFFERARBPROC wglDestroyPbufferARB = NULL;

static PFNWGLBINDTEXIMAGEARBPROC wglBindTexImageARB = NULL;
static PFNWGLRELEASETEXIMAGEARBPROC wglReleaseTexImageARB = NULL;

void* getProcAddress(char* name)
{
   void *ptr= NULL;

   ptr= (void*)wglGetProcAddress(name);

   if (!ptr) {
	   char err[256];
	   sprintf(err, "GL extension not available: %s\n", name);
	   MessageBox(NULL, err, "Error", MB_OK);
   }

   return ptr;
}

bool gfxInitExtensions()
{
	//printf("extensions:\n");
	char *ext= (char*)glGetString(GL_EXTENSIONS);
	//MessageBox(NULL, ext, "GL Extensions", MB_OK);

	while (*ext!=0)
   {
      char *temp=ext;
      while (*temp!=0 && *temp!=32) temp++;
      if (*temp==0)
      {
         //printf("%s\n", ext);
         ext=temp;
      }
      else
      {
         *temp=0;
         //printf("%s\n", ext);
         ext= temp+1;
      }
   }

// TODO
// check available extensions and drop out on error!
//	glDrawRangeElementsARB= (PFNGLDRAWRANGEELEMENTSPROC) getProcAddress("glDrawRangeElementsARB");
//	glCreateShader =   (PFNGLCREATESHADERPROC)  getProcAddress("glCreateShader");
	glVertexAttrib4f =   (PFNGLVERTEXATTRIB4FPROC)  getProcAddress("glVertexAttrib4fARB");
	glVertexAttribPointer =   (PFNGLVERTEXATTRIBPOINTERPROC)  getProcAddress("glVertexAttribPointerARB");
	glEnableVertexAttribArray = (PFNGLENABLEVERTEXATTRIBARRAYPROC) getProcAddress("glEnableVertexAttribArrayARB");
	glDisableVertexAttribArray = (PFNGLDISABLEVERTEXATTRIBARRAYPROC) getProcAddress("glDisableVertexAttribArrayARB");

	// multi-texturing
	glClientActiveTextureARB= (PFNGLCLIENTACTIVETEXTUREPROC) getProcAddress("glClientActiveTextureARB");
//	glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC) getProcAddress("glActiveTextureARB");
	glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC) getProcAddress("glActiveTexture");
	glMultiTexCoord2fARB= (PFNGLMULTITEXCOORD2FPROC) getProcAddress("glMultiTexCoord2fARB");
	glMultiTexCoord3fARB= (PFNGLMULTITEXCOORD3FPROC) getProcAddress("glMultiTexCoord3fARB");

	// vertex buffer objects
	glGenBuffersARB =    (PFNGLGENBUFFERSARBPROC)    getProcAddress("glGenBuffersARB");
	glBindBufferARB =    (PFNGLBINDBUFFERARBPROC)    getProcAddress("glBindBufferARB");
	glBufferDataARB =    (PFNGLBUFFERDATAARBPROC)    getProcAddress("glBufferDataARB");
	glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC) getProcAddress("glDeleteBuffersARB");
	glMapBufferARB =     (PFNGLMAPBUFFERPROC)        getProcAddress("glMapBufferARB");
	glUnmapBufferARB =   (PFNGLUNMAPBUFFERPROC)      getProcAddress("glUnmapBufferARB");

	// ARB_vertex_program + ARB_fragment_program
	glGenProgramsARB				= (PFNGLGENPROGRAMSARBPROC) getProcAddress("glGenProgramsARB");
	glBindProgramARB				= (PFNGLBINDPROGRAMARBPROC) getProcAddress("glBindProgramARB");
	glProgramStringARB				= (PFNGLPROGRAMSTRINGARBPROC) getProcAddress("glProgramStringARB");
	glDeleteProgramsARB				= (PFNGLDELETEPROGRAMSARBPROC) getProcAddress("glDeleteProgramsARB");
	glProgramLocalParameter4fARB	= (PFNGLPROGRAMLOCALPARAMETER4FARBPROC) getProcAddress("glProgramLocalParameter4fARB");
	glProgramLocalParameter4fvARB	= (PFNGLPROGRAMLOCALPARAMETER4FVARBPROC) getProcAddress("glProgramLocalParameter4fvARB");
	glProgramEnvParameter4fARB		= (PFNGLPROGRAMENVPARAMETER4FARBPROC) getProcAddress("glProgramEnvParameter4fARB");
	glProgramEnvParameter4fvARB		= (PFNGLPROGRAMENVPARAMETER4FVARBPROC) getProcAddress("glProgramEnvParameter4fvARB");

	// GLSL shader stuff
	glCreateShaderObjectARB =   (PFNGLCREATESHADEROBJECTARBPROC)  getProcAddress("glCreateShaderObjectARB");
	glShaderSourceARB =         (PFNGLSHADERSOURCEARBPROC)        getProcAddress("glShaderSourceARB");
	glCompileShaderARB =        (PFNGLCOMPILESHADERARBPROC)       getProcAddress("glCompileShaderARB");
	glCreateProgramObjectARB =  (PFNGLCREATEPROGRAMOBJECTARBPROC) getProcAddress("glCreateProgramObjectARB");
	glAttachObjectARB =         (PFNGLATTACHOBJECTARBPROC)        getProcAddress("glAttachObjectARB");
	glLinkProgramARB =          (PFNGLLINKPROGRAMARBPROC)         getProcAddress("glLinkProgramARB");
	glUseProgramObjectARB =     (PFNGLUSEPROGRAMOBJECTARBPROC)    getProcAddress("glUseProgramObjectARB");
	glGetObjectParameterfvARB = (PFNGLGETOBJECTPARAMETERFVARBPROC)getProcAddress("glGetObjectParameterfvARB");
	glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)getProcAddress("glGetObjectParameterivARB");
	glGetUniformLocationARB=    (PFNGLGETUNIFORMLOCATIONARBPROC)  getProcAddress("glGetUniformLocationARB");
	glUniform1iARB =            (PFNGLUNIFORM1IARBPROC)           getProcAddress("glUniform1iARB");
	glUniform1fARB =            (PFNGLUNIFORM1FARBPROC)           getProcAddress("glUniform1fARB");
	glUniform2fARB =            (PFNGLUNIFORM2FARBPROC)           getProcAddress("glUniform2fARB");
	glUniform3fARB =            (PFNGLUNIFORM3FARBPROC)           getProcAddress("glUniform3fARB");
	glUniform4fARB =            (PFNGLUNIFORM4FARBPROC)           getProcAddress("glUniform4fARB");
	glUniform1fvARB =           (PFNGLUNIFORM1FVARBPROC)          getProcAddress("glUniform1fvARB");
	glUniform2fvARB =           (PFNGLUNIFORM2FVARBPROC)          getProcAddress("glUniform2fvARB");
	glUniform3fvARB =           (PFNGLUNIFORM3FVARBPROC)          getProcAddress("glUniform3fvARB");
	glUniform4fvARB =           (PFNGLUNIFORM4FVARBPROC)          getProcAddress("glUniform4fvARB");

	glGetInfoLogARB=            (PFNGLGETINFOLOGARBPROC)          getProcAddress("glGetInfoLogARB");

	/*
	glGenFramebuffersEXT= (PFNGLGENFRAMEBUFFERSEXTPROC)getProcAddress("glGenFramebuffersEXT");
	glBindFramebufferEXT= (PFNGLBINDFRAMEBUFFEREXTPROC)getProcAddress("glBindFramebufferEXT");
	glGenRenderbuffersEXT= (PFNGLGENRENDERBUFFERSEXTPROC)getProcAddress("glGenRenderbuffersEXT");
	glRenderbufferStorageEXT= (PFNGLRENDERBUFFERSTORAGEEXTPROC)getProcAddress("glRenderbufferStorageEXT");
	glFramebufferRenderbufferEXT= (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)getProcAddress("glFramebufferRenderbufferEXT");
	glFramebufferTexture2DEXT= (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)getProcAddress("glFramebufferTexture2DEXT");
	glCheckFramebufferStatusEXT= (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC)getProcAddress("glCheckFramebufferStatusEXT");
	glBindRenderbufferEXT= (PFNGLBINDRENDERBUFFEREXTPROC)getProcAddress("glBindRenderbufferEXT");
	*/

	wglChoosePixelFormatARB =	(PFNWGLCHOOSEPIXELFORMATARBPROC)	getProcAddress("wglChoosePixelFormatARB");
	wglCreatePbufferARB =		(PFNWGLCREATEPBUFFERARBPROC)		getProcAddress("wglCreatePbufferARB");
	wglGetPbufferDCARB =		(PFNWGLGETPBUFFERDCARBPROC)			getProcAddress("wglGetPbufferDCARB");
	wglReleasePbufferDCARB =	(PFNWGLRELEASEPBUFFERDCARBPROC)		getProcAddress("wglReleasePbufferDCARB");
	wglDestroyPbufferARB =		(PFNWGLDESTROYPBUFFERARBPROC)		getProcAddress("wglDestroyPbufferARB");

	wglBindTexImageARB = (PFNWGLBINDTEXIMAGEARBPROC) getProcAddress("wglBindTexImageARB");
	wglReleaseTexImageARB = (PFNWGLRELEASETEXIMAGEARBPROC) getProcAddress("wglReleaseTexImageARB");
/*
	glActiveStencilFaceEXT=     (PFNGLACTIVESTENCILFACEEXTPROC)   getProcAddress("glActiveStencilFaceEXT");
	glStencilOpSeparateATI=     (PFNGLSTENCILOPSEPARATEATIPROC)   getProcAddress("glStencilOpSeparateATI");
	glStencilFuncSeparateATI=   (PFNGLSTENCILFUNCSEPARATEATIPROC) getProcAddress("glStencilFuncSeparateATI");
*/
   return true;
}


HGLRC gfxCreateGLContext( HDC hdc, bool need_window )
{
	PIXELFORMATDESCRIPTOR pfd;
    int  pixelFormat;
	HGLRC hglrc;

	memset(&pfd,0,sizeof(pfd));
	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 32;
	pfd.cDepthBits = 24;
	pfd.cStencilBits = 8;
	pfd.cAlphaBits = 8;
	pfd.iLayerType = PFD_MAIN_PLANE;

	if (need_window) pfd.dwFlags |= PFD_DRAW_TO_WINDOW;

	pixelFormat = ChoosePixelFormat(hdc, &pfd);
	if (pixelFormat)
	{
        SetPixelFormat(hdc, pixelFormat, &pfd);
		{
            hglrc = wglCreateContext(hdc);	
            if (hglrc) 
			{
                if (wglMakeCurrent(hdc, hglrc))
				{
					//wglMakeCurrent(0, 0);
					return hglrc;
				}
				wglDeleteContext(hglrc);
            }
		}
	}

	return 0;
}

void gfxSetContext(HGLRC hglrc, HDC dc)
{
	currentHGLRC = hglrc;
	currentDC = dc;
	wglMakeCurrent(currentDC, currentHGLRC);
}

void gfxDeleteContext()
{
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(currentHGLRC);
}

void gfxFlip()
{
	SwapBuffers(currentDC);
}

//--------------------------------------------
// PBuffer
//--------------------------------------------

struct RenderTargetData {
	HDC hdc;
	HPBUFFERARB pbuffer;
	HGLRC context;
	GLuint glTexID;
	
	HDC old_hdc;
	HGLRC old_context;
};

void gfxCreateRenderTarget(RenderTarget* rt, uint width, uint height, uint flags) {	
	assertCheck(rt, "RenderTarget is NULL!");

	HDC old_hdc = wglGetCurrentDC();
	HGLRC old_context = wglGetCurrentContext();
	
	int a = 0;
	int attrib[128];
	memset(attrib, 0, sizeof(int)*128);
	
	int componentBits = (flags & RENDERTARGET_FLOAT ? 32 : 8);

	if(flags & RENDERTARGET_RGBA) {
		attrib[a++] = WGL_RED_BITS_ARB;
		attrib[a++] = componentBits;
		attrib[a++] = WGL_GREEN_BITS_ARB;
		attrib[a++] = componentBits;
		attrib[a++] = WGL_BLUE_BITS_ARB;
		attrib[a++] = componentBits;
		attrib[a++] = WGL_ALPHA_BITS_ARB;
		attrib[a++] = componentBits;
	}
	
	if(flags & RENDERTARGET_DEPTH) {
		attrib[a++] = WGL_DEPTH_BITS_ARB;
		attrib[a++] = 24;
	}
	
	if(flags & RENDERTARGET_STENCIL) {
		attrib[a++] = WGL_STENCIL_BITS_ARB;
		attrib[a++] = 8;
	}
	
	if(flags & RENDERTARGET_FLOAT) {
		attrib[a++] = WGL_FLOAT_COMPONENTS_NV;
		attrib[a++] = GL_TRUE;
	}
	
	if(flags & RENDERTARGET_MULTISAMPLE_2 || flags & RENDERTARGET_MULTISAMPLE_4) {
		attrib[a++] = WGL_SAMPLE_BUFFERS_ARB;
		attrib[a++] = GL_TRUE;
		attrib[a++] = WGL_SAMPLES_ARB;
		attrib[a++] = (flags & RENDERTARGET_MULTISAMPLE_2 ? 2 : 4);
	}
	
	attrib[a++] = WGL_DRAW_TO_PBUFFER_ARB;
	attrib[a++] = GL_TRUE;
	attrib[a++] = WGL_SUPPORT_OPENGL_ARB;
	attrib[a++] = GL_TRUE;
	attrib[a++] = WGL_BIND_TO_TEXTURE_RGBA_ARB;
	attrib[a++] = GL_TRUE;
	attrib[a++] = WGL_ACCELERATION_ARB;
	attrib[a++] = WGL_FULL_ACCELERATION_ARB;

	attrib[a++] = WGL_PIXEL_TYPE_ARB;
	attrib[a++] = WGL_TYPE_RGBA_ARB;
	attrib[a++] = 0;
	
	HDC hdc;
	HPBUFFERARB pbuffer;
	HGLRC context;	
		
	int format;
	unsigned int count;
	wglChoosePixelFormatARB(old_hdc,&attrib[0],NULL,1,&format,&count);
	assertCheck(count != 0, "wglChoosePixelFormatARB() failed");
	
//	int pattrib[] = { 0 };
	int pattrib[] = {
			WGL_TEXTURE_FORMAT_ARB,  WGL_TEXTURE_RGBA_ARB,
			WGL_TEXTURE_TARGET_ARB,  WGL_TEXTURE_2D_ARB,
			0, 0
	};

	pbuffer = wglCreatePbufferARB(old_hdc,format,width,height,pattrib);
	assertCheck(pbuffer, "wglCreatePbufferARB() failed");
	
	hdc = wglGetPbufferDCARB(pbuffer);
	assertCheck(hdc, "wglGetPbufferDCARB() failed");
	
	context = wglCreateContext(hdc);
	assertCheck(context, "wglCreateContext() failed");
	
	assertCheck(wglShareLists(old_context,context), "wglShareLists() failed");

	RenderTargetData* data = new RenderTargetData;
	data->hdc = hdc;
	data->pbuffer = pbuffer;
	data->context = context;
	
	data->old_hdc = old_hdc;
	data->old_context = old_context;

	wglMakeCurrent(hdc, context);
	
	// set default states
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	//glFrontFace(GL_CW);
	glEnable(GL_CULL_FACE);

	wglMakeCurrent(old_hdc, old_context);

	glGenTextures(1, &data->glTexID);
	glBindTexture(GL_TEXTURE_2D, data->glTexID);
	// TODO: setup filtering
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	rt->data = data;
	rt->width = width;
	rt->height = height;
}

void gfxReleaseRenderTarget(RenderTarget* rt) {
	assertCheck(rt, "RenderTarget is NULL!");

	wglDeleteContext(rt->data->context);
	wglReleasePbufferDCARB(rt->data->pbuffer, rt->data->hdc);
	wglDestroyPbufferARB(rt->data->pbuffer);
	wglMakeCurrent(rt->data->hdc, rt->data->context);

	if (rt->data) { 
		delete rt->data;
		rt->data = NULL;
	}
}

void gfxEnableRenderTarget(RenderTarget* rt) {
	assertCheck(rt, "RenderTarget is NULL!");

	rt->data->old_hdc = wglGetCurrentDC();
	rt->data->old_context = wglGetCurrentContext();
	
	assertCheck(wglMakeCurrent(rt->data->hdc, rt->data->context), "wglMakeCurrent() failed!");
}

void gfxDisableRenderTarget(RenderTarget* rt) {
	assertCheck(rt, "RenderTarget is NULL!");
	assertCheck(wglMakeCurrent(rt->data->old_hdc, rt->data->old_context), "wglMakeCurrent() failed!");
}

/*
void R_SetRenderTarget(uint texUnit, RenderTarget* rt) {
	glActiveTexture(GL_TEXTURE0 + texUnit);
	if (!rt) wglReleaseTexImageARB(rt->data->pbuffer, WGL_FRONT_LEFT_ARB);
	else {
		glBindTexture(GL_TEXTURE_2D, rt->data->glTexID);
		wglBindTexImageARB(rt->data->pbuffer, WGL_FRONT_LEFT_ARB);
	}
}*/

void gfxBindRenderTarget(uint texUnit, RenderTarget* rt) {
	glActiveTextureARB(GL_TEXTURE0 + texUnit);
	glBindTexture(GL_TEXTURE_2D, rt->data->glTexID);
	wglBindTexImageARB(rt->data->pbuffer, WGL_FRONT_LEFT_ARB);
}

void gfxUnbindRenderTarget(uint texUnit, RenderTarget* rt) {
	glActiveTextureARB(GL_TEXTURE0 + texUnit);
	wglReleaseTexImageARB(rt->data->pbuffer, WGL_FRONT_LEFT_ARB);
}