#include "TexManager.h"
#include "RenderAPI.h"
#include <string.h>

extern "C" u8 *stbi_load(char const *filename, int *x, int *y, int *comp, int req_comp);
extern "C" void stbi_image_free(void *retval_from_stbi_load);

TextureManager::TextureDesc* TextureManager::FindTexture(const char* filename) {
	for (size_t t=0; t < m_tex_count; t++) {
		if (strcmp(m_textures[t].filename, filename) == 0) return &m_textures[t];
	}
	return NULL;
}

void flipImage(u8* data, int w, int h) {
	int pitch = w*4;
	u8* tmpScanline = new u8[pitch];
	for (int y=0; y < w/2; y++) {
		u8* dstScanline = data + y * pitch;
		u8* srcScanline = data + (h-y-1) * pitch;
		memcpy(tmpScanline, dstScanline, pitch);
		memcpy(dstScanline, srcScanline, pitch);
		memcpy(srcScanline, tmpScanline, pitch);
	}
	delete[] tmpScanline;
}

u32 TextureManager::GetTexture(const char* filename, bool mipMaps) {
	TextureDesc* tx = FindTexture(filename);
	if (tx) {
		tx->refCount++;
		return tx->handle;
	}
	TextureDesc tex;
	strcpy(tex.filename, filename);
	tex.refCount = 1;
	tex.mipMaps = mipMaps;

	int x,y,n;
	unsigned char *data = stbi_load(filename, &x, &y, &n, 4);

	if (!data) return 0;

	flipImage(data, x, y);

	tex.handle = gfxUploadTexture2D(data, x, y, (u32)mipMaps);

	stbi_image_free(data);

	m_textures[m_tex_count++] = tex;

	return tex.handle;
}

void TextureManager::FreeTextures() {
	for (size_t t=0; t < m_tex_count; t++) {
		gfxReleaseTexture(m_textures[t].handle);
	}
	m_tex_count = 0;
}
