#include "Defs.h"
#include <algorithm>
#include <windows.h>

#include <exception>

namespace std {
	void (*_Raise_handler)(const exception&) = NULL;
	void __cdecl _Throw(const exception&) {}
}

//----------------------------------------------------------------------------
// Strings
//----------------------------------------------------------------------------
namespace StringUtils
{

void toLower(std::string& s)
{
	transform(s.begin(), s.end(), s.begin(), tolower);
}

void toUpper(std::string& s)
{
	transform(s.begin(), s.end(), s.begin(), toupper);
}

void replace(std::string& s, const std::string& find, const std::string& replacement)
{
	if(find.empty()) return;

	size_t replace_inc = replacement.length();

	for(size_t loc = s.find(find); loc != std::string::npos; loc = s.find(find, loc+replace_inc))	
		s.replace(loc, find.length(), replacement);
} 

}

//----------------------------------------------------------------------------
// Timer
//----------------------------------------------------------------------------

Timer::Timer()
{
	liTimeCurrent = 0;
	liTimeFreq = 0;
	liTimePrev = 0;
	liTimeStart = 0;

	timeCurrent = 0;
	timePrev = 0;
	timeStart = 0;
	timeDelta = 0;

	if(!QueryPerformanceFrequency((LARGE_INTEGER*)&liTimeFreq))
	{
		// fix: exit app and tell the user that he doesnt have requested hardware;
	}
}

bool Timer::Init()
{
	QueryPerformanceCounter((LARGE_INTEGER*)&liTimeStart);
	timeStart = (float)liTimeStart/liTimeFreq;
	return true;
}

void Timer::Update()
{
	liTimePrev = liTimeCurrent;
	QueryPerformanceCounter((LARGE_INTEGER*)&liTimeCurrent);
	liTimeCurrent -= liTimeStart;

	timePrev = timeCurrent;
	timeCurrent = (float)liTimeCurrent/liTimeFreq;

	timeDelta = timeCurrent - timePrev;
}
