#pragma once
#include <assert.h>

#define _HAS_EXCEPTIONS 0
#define _STATIC_CPPLIB 1

#include <string>

typedef short i16;
typedef int i32;
typedef __int64 i64;
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned __int64 u64;

typedef unsigned int uint;

typedef unsigned int UID;

#define ILINE __forceinline

#define SAFE_DELETE(p)       { if(p) {delete (p); (p)=NULL;} }
#define SAFE_DELETE_ARRAY(p) { if(p) {delete[] (p); (p)=NULL;} }

#define assertFail(msg) _assert(__FUNCTION__"() - "msg, __FILE__, __LINE__)
#define assertCheck(expr, msg) if(!(expr)) assertFail(msg);

inline void osDebugBreak() { __asm int 3 }

extern void errorMessage(const char *format, ...);
extern void logMessage(const char *format, ...);

// number of set bits in the mask
//
// from Steve Baker's "Cute Code" collection
// http://web2.airmail.net/sjbaker1/software/cute_code.html
inline int CountBitsInMask(u32 mask)
{
	mask = (mask & 0x55555555) + ((mask & 0xaaaaaaaa) >> 1);
	mask = (mask & 0x33333333) + ((mask & 0xcccccccc) >> 2);
	mask = (mask & 0x0f0f0f0f) + ((mask & 0xf0f0f0f0) >> 4);
	mask = (mask & 0x00ff00ff) + ((mask & 0xff00ff00) >> 8);
	mask = (mask & 0x0000ffff) + (mask >> 16);
	return mask;
}

namespace StringUtils
{
	void toLower(std::string& s);
	void toUpper(std::string& s);
	void replace(std::string& s, const std::string& find, const std::string& replacement);

	// String parsing functions
	inline bool isBlankChar(char c) { return (c == ' ' || c == '\t'); }
	inline bool isNewLineChar(char c) { return (c == '\n' || c == '\r'); }
	inline char* getNextToken(char* buffer = NULL)
	{
		const char v[] = { ' ', '\n', 9, 13, 0 };
		return strtok(buffer,v);
	}
	inline char* getNextString(char* buffer = NULL)
	{
		const char v[] = { '\n', 9, 13, 0 };
		return strtok(buffer,v);
	}
	inline char* skipLine(char* buffer) {
		while(!isNewLineChar(*buffer++)) {}
		while(isNewLineChar(*buffer++)) {}
		return buffer-1;
	}
}

//----------------------------------------------------------------------------
// Timer
//----------------------------------------------------------------------------
class Timer
{
public:
	Timer();

	bool Init();

	// time in sec
	inline float GetCurTime() { return timeCurrent; }
	inline float GetPrevTime() { return timePrev; }
	inline float GetDeltaTime() { return timeDelta; }
	inline float GetStartTime() { return timeStart; }

	void Update();

protected:
	float timeCurrent;
	float timePrev;
	float timeDelta;
	float timeStart;
private:
	u64 liTimeFreq;
	u64 liTimePrev;
	u64 liTimeCurrent;
	u64 liTimeStart;
};
