#include "Math.h"
	
Vec3f Quat::RotateVector(const Vec3f &v)
{
	Vec3f ret;
	Quat tmp, inv, final;

	inv.x = -x; inv.y = -y;
	inv.z = -z; inv.w =  w;

	inv.Normalize();

	tmp = (*this) * v;
	final = tmp * inv;

	ret.x = final.x;
	ret.y = final.y;
	ret.z = final.z;
	return ret;
}

void Quat::Slerp(const Quat& from, const Quat& to, float t)
{
	Quat to1;
	float omega, cosO, sinO;
	float scale0, scale1;

	cosO = from.x*to.x + from.y*to.y + from.z*to.z + from.w*to.w;

	if(cosO < 0.0f)
	{
		cosO = -cosO;
		to1 = Quat(-to.x, -to.y, -to.z, -to.w);
	}
	else to1 = to;

	if((1.0f - cosO) > 1.0E-06)
	{
		omega = (float) acos(cosO);
		sinO = (float) sin(omega);
		scale0 = (float) sin((1.0f - t) * omega) / sinO;
		scale1 = (float) sin(t * omega) / sinO;
	}
	else
	{
		scale0 = 1.0f - t;
		scale1 = t;
	}

	*this = from * scale0 + to1 * scale1;
}

void Mtx4f::Invert() {
    double a,b,c,d,e,f,g,h,i,j,k,l;
    register double dW;

    a = m[0].x; b = m[0].y; c = m[0].z;
    d = m[1].x; e = m[1].y; f = m[1].z;
    g = m[2].x; h = m[2].y; i = m[2].z;
    j = m[3].x; k = m[3].y; l = m[3].z;

    dW = 1.0 / (a*(e*i - f*h) - (b*(d*i - f*g) + c*(e*g - d*h)));

    m[0].x = (float)((e*i - f*h) * dW);
    m[0].y = (float)((c*h - b*i) * dW);
    m[0].z = (float)((b*f - c*e) * dW);

    m[1].x = (float)((f*g - d*i) * dW);
    m[1].y = (float)((a*i - c*g) * dW);
    m[1].z = (float)((c*d - a*f) * dW);

    m[2].x = (float)((d*h - e*g) * dW);
    m[2].y = (float)((b*g - a*h) * dW);
    m[2].z = (float)((a*e - b*d) * dW);

    m[3].x = (float)((e*(g*l - i*j) + f*(h*j - g*k) - d*(h*l - i*k)) * dW);
    m[3].y = (float)((a*(h*l - i*k) + b*(i*j - g*l) + c*(g*k - h*j)) * dW);
    m[3].z = (float)((b*(d*l - f*j) + c*(e*j - d*k) - a*(e*l - f*k)) * dW);
}

Mtx4f Mtx4f::GetInverted() const
{
   Mtx4f n;

   float det= m[0].x*(m[1].y*m[2].z - m[2].y*m[1].z) + m[1].x*(m[2].y*m[0].z - m[0].y*m[2].z) + m[2].x*(m[0].y*m[1].z - m[1].y*m[0].z);

   // matrix not invertible?
   //   if (det==0) return ::identity();

   float inv= 1.0f / det;

   // generate adjoint matrix
   n.m[0].x= (m[1].y*m[2].z - m[2].y*m[1].z)*inv;
   n.m[0].y= (m[2].y*m[0].z - m[0].y*m[2].z)*inv;
   n.m[0].z= (m[0].y*m[1].z - m[1].y*m[0].z)*inv;
   n.m[0].w= 0.0f;

   n.m[1].x= (m[2].x*m[1].z - m[1].x*m[2].z)*inv;
   n.m[1].y= (m[0].x*m[2].z - m[2].x*m[0].z)*inv;
   n.m[1].z= (m[1].x*m[0].z - m[0].x*m[1].z)*inv;
   n.m[1].w= 0.0f;

   n.m[2].x= (m[1].x*m[2].y - m[2].x*m[1].y)*inv;
   n.m[2].y= (m[2].x*m[0].y - m[0].x*m[2].y)*inv;
   n.m[2].z= (m[0].x*m[1].y - m[1].x*m[0].y)*inv;
   n.m[2].w= 0.0f;

   // new translation vector = negative transposed rot/scale mat *  old translation vector
   n.m[3].x = -(n.m[0].x*m[3].x) - (n.m[1].x*m[3].y) - (n.m[2].x*m[3].z);
   n.m[3].y = -(n.m[0].y*m[3].x) - (n.m[1].y*m[3].y) - (n.m[2].y*m[3].z);
   n.m[3].z = -(n.m[0].z*m[3].x) - (n.m[1].z*m[3].y) - (n.m[2].z*m[3].z);
   n.m[3].w= 1.0f;

  return n;
}
void Mtx4f::LookAt(const Vec3f& pos, const Vec3f& target, const Vec3f& up) {
	Vec3f forward = pos - target;
	forward.Normalize();
	Vec3f right = CrossProduct(up, forward);
	right.Normalize();
	Vec3f newUp = CrossProduct(forward, right);
	newUp.Normalize();

	Mtx4f rot;
	rot.SetRow0(Vec4f(right, 0));
	rot.SetRow1(Vec4f(newUp, 0));
	rot.SetRow2(Vec4f(forward, 0));
	rot.SetRow3(Vec4f(0,0,0,1));

	Mtx4f trans;
	trans.Identity();
	trans.SetTranslation(-pos);

	*this = rot * trans;
//	*this = trans * rot;
}

Quat operator*(Quat a, Quat b) {
	Quat c;
	c.w = a.w*b.w - a.x*b.x - a.y*b.y - a.z*b.z; 
	c.x = a.w*b.x + a.x*b.w + a.y*b.z - a.z*b.y; 
	c.y = a.w*b.y - a.x*b.z + a.y*b.w + a.z*b.x; 
	c.z = a.w*b.z + a.x*b.y - a.y*b.x + a.z*b.w; 
	return c;
}
/*
  out[W] = (qa[W] * qb[W]) - (qa[X] * qb[X]) - (qa[Y] * qb[Y]) - (qa[Z] * qb[Z]);
  out[X] = (qa[X] * qb[W]) + (qa[W] * qb[X]) + (qa[Y] * qb[Z]) - (qa[Z] * qb[Y]);
  out[Y] = (qa[Y] * qb[W]) + (qa[W] * qb[Y]) + (qa[Z] * qb[X]) - (qa[X] * qb[Z]);
  out[Z] = (qa[Z] * qb[W]) + (qa[W] * qb[Z]) + (qa[X] * qb[Y]) - (qa[Y] * qb[X]);
*/

void Mtx4f::FromQuat(Quat q) {
	Vec3f x = q.xdir();
	Vec3f y = q.ydir();
	Vec3f z = q.zdir();
	m[0].x = x.x;
	m[0].y = x.y;
	m[0].z = x.z;
	m[1].x = y.x;
	m[1].y = y.y;
	m[1].z = y.z;
	m[2].x = z.x;
	m[2].y = z.y;
	m[2].z = z.z;
}

Mtx4f Mtx4f::TranslationMatrix(const Vec3f& t) {
	Mtx4f trans;
	trans.Set(1.0f, 0.0f, 0.0f, t.x,
			  0.0f, 1.0f, 0.0f, t.y,
			  0.0f, 0.0f, 1.0f, t.z,
			  0.0f, 0.0f, 0.0f, 1.0f);
	return trans;
}

Mtx4f Mtx4f::RotationMatrixX(float angle) {
	float c = cosf(angle);
	float s = sinf(angle);
	Mtx4f rota;
	rota.Set(1, 0, 0, 0,
			 0, c,-s, 0,
			 0, s, c, 0,
			 0, 0, 0, 1);
	return rota;
}

Mtx4f Mtx4f::RotationMatrixY(float angle) {
	float c = cosf(angle);
	float s = sinf(angle);
	Mtx4f rota;
	rota.Set(c, 0, s, 0,
			 0, 1, 0, 0,
			 -s, 0, c, 0,
			 0, 0, 0, 1);
	return rota;
}

Mtx4f Mtx4f::RotationMatrixZ(float angle) {
	float c = cosf(angle);
	float s = sinf(angle);
	Mtx4f rota;
	rota.Set(c,-s, 0, 0,
			 s, c, 0, 0,
			 0, 0, 1, 0,
			 0, 0, 0, 1);
	return rota;
}

void Mtx4f::Translate(const Vec3f& t) {
	*this = TranslationMatrix(t) * *this;
}

Mtx4f Mtx4f::GetOrientation() {
	Mtx4f rota = *this;
	rota.SetTranslation(Vec3f(0,0,0));
	rota.SetRow3(Vec4f(0,0,0,1));
	return rota;
}
